/*
 * Copyright (c) 2011 Miyakawa Taku
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package org.bitbucket.miyakawa_taku.hadoop_job_builder.mapred;

import org.bitbucket.miyakawa_taku.hadoop_job_builder.internal.Commons;

import java.util.List;
import java.util.ArrayList;
import java.lang.reflect.Type;
import java.lang.reflect.ParameterizedType;
import java.net.URI;
import java.net.URISyntaxException;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Partitioner;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.TextInputFormat;
import org.apache.hadoop.mapred.KeyValueTextInputFormat;
import org.apache.hadoop.mapred.SequenceFileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.apache.hadoop.mapred.SequenceFileOutputFormat;
import org.apache.hadoop.mapred.lib.IdentityMapper;
import org.apache.hadoop.mapred.lib.IdentityReducer;
import org.apache.hadoop.io.RawComparator;
import org.apache.hadoop.filecache.DistributedCache;

/**
 * Type-safe builder of {@linkplain JobConf a job configuration}
 * of {@linkplain org.apache.hadoop.mapred the old API of Hadoop MapReduce}.
 *
 * <p>This builder allows comple-time type checking
 * between {@linkplain JobConf#setMapperClass a mapper},
 * {@linkplain JobConf#setReducerClass a reducer},
 * {@linkplain JobConf#setCombinerClass a combiner},
 * {@linkplain JobConf#setPartitionerClass a partitioner},
 * {@linkplain RawComparator comparators}, keys and values.
 * This also provides convenient methods to set up {@linkplain DistributedCache}
 * and several common {@linkplain JobConf#setInputFormat input formats}
 * and {@linkplain JobConf#setOutputFormat output formats}.</p>
 *
 * <h3>Example</h3>
 *
 * <pre id="simplisticexample">
 * JobConf jobConf = JobConfBuilder.of( new WordCountMapper() , new WordCountReducer() )
 *     .jobName( "word-count" )
 *     .detectJar()
 *     .detectKeyValue()
 *     .combiner( new WordCountReducer() )
 *     .inputTextFrom( "word-count/input" )
 *     .outputSequenceFileOn( "word-count/output" )
 *     .buildJobConf( getConf() );
 * // launches the job
 * JobClient.runJob( jobConf );
 * </pre>
 *
 * <h3>Equivalent code using {@linkplain org.apache.hadoop.mapred the old MapReduce API}</h3>
 *
 * <pre>
 * JobConf jobConf = new JobConf( getConf() );
 * jobConf.setJobName( "word-count" );
 * jobConf.setJarByClass( getClass() );
 * jobConf.setMapperClass( WordCountMapper.class );
 * jobConf.setReducerClass( WordCountReducer.class );
 * jobConf.setMapOutputKeyClass( Text.class );
 * jobConf.setMapOutputValueClass( IntWritable.class );
 * jobConf.setOutputKeyClass( Text.class );
 * jobConf.setOutputValueClass( IntWritable.class );
 * jobConf.setCombinerClass( WordCountReducer.class );
 * jobConf.setInputFormat( TextInputFormat.class );
 * FileInputFormat.addInputPath( jobConf , new Path( "word-count/input" ) );
 * jobConf.setOutputFormat( SequenceFileOutputFormat.class );
 * FileOutputFormat.setOutputPath( jobConf , new Path( "word-count/output" ) );
 * // launches the job
 * JobClient.runJob( jobConf );
 * </pre>
 *
 * <h3>Given these mapper/reducer classes</h3>
 *
 * <pre>
 * public static class WordCountMapper
 *         implements Mapper&lt; LongWritable , Text , Text , IntWritable &gt; {
 *     &#64;Override public void map(...) { ... }
 *     ...
 * }
 *
 * public static class WordCountReducer
 *         implements Reducer&lt; Text , IntWritable , Text , IntWritable &gt; {
 *     &#64;Override public void reduce(...) { ... }
 *     ...
 * }
 * </pre>
 *
 * <h3>Additional configuration</h3>
 *
 * <p>You can of cource set up a job configuration additionally
 * for any purpose.</p>
 *
 * <pre>
 * JobConf jobConf = JobConfBuilder.of( new WordCountMapper() , new WordCountReducer() )
 *     .jobName( "word-count" )
 *     .detectJar()
 *     .detectKeyValue()
 *     .combiner( new WordCountReducer() )
 *     .inputTextFrom( "word-count/input" )
 *     .outputSequenceFileOn( "word-count/output" )
 *     .buildJobConf( getConf() );
 * // additional configuration
 * SequenceFileOutputFormat.setOutputCompressionType( jobConf , BLOCK );
 * // launches the job
 * JobClient.runJob( jobConf );
 * </pre>
 *
 * <h3>Index</h3>
 *
 * <dl>
 *   <dt>Set up the mapper and the reducer</dt>
 *   <dd>{@link #of(Mapper,Reducer)}</dd>
 *   <dd>{@link #of(Mapper)}</dd>
 *
 *   <dt>Set up the keys and values</dt>
 *   <dd>{@link #mapOutputKeyValue(Class,Class)}</dd>
 *   <dd>{@link #outputKeyValue(Class,Class)}</dd>
 *   <dd>{@link #detectKeyValue}</dd>
 *   <dd>{@link #detectMapOutputKeyValue}</dd>
 *   <dd>{@link #detectOutputKeyValue}</dd>
 *
 *   <dt>Set up the combiner</dt>
 *   <dd>{@link #combiner}</dd>
 *
 *   <dt>Set up the partitioner</dt>
 *   <dd>{@link #partitioner}</dd>
 *
 *   <dt>Set up the comparators</dt>
 *   <dd>{@link #sortComparator}</dd>
 *   <dd>{@link #groupingComparator}</dd>
 *
 *   <dt>Set up distributed cache</dt>
 *   <dd>{@link #cacheFile}</dd>
 *   <dd>{@link #cacheFileWithSymlink}</dd>
 *   <dd>{@link #cacheArchive}</dd>
 *   <dd>{@link #cacheArchiveWithSymlink}</dd>
 *
 *   <dt>Set up the job name</dt>
 *   <dd>{@link #jobName}</dd>
 *
 *   <dt>Set up the jar file name</dt>
 *   <dd>{@link #detectJar}</dd>
 *   <dd>{@link #jarByClass}</dd>
 *
 *   <dt>Set up the working directory</dt>
 *   <dd>{@link #workingDirectory(Path)}</dd>
 *   <dd>{@link #workingDirectory(String)}</dd>
 *
 *   <dt>Set up the number of reduce tasks</dt>
 *   <dd>{@link #numReduceTasks}</dd>
 *
 *   <dt>Set up the input format</dt>
 *   <dd>{@link #inputKeyValueTextFrom(Path...)}</dd>
 *   <dd>{@link #inputKeyValueTextFrom(String...)}</dd>
 *   <dd>{@link #inputTextFrom(Path...)}</dd>
 *   <dd>{@link #inputTextFrom(String...)}</dd>
 *   <dd>{@link #inputSequenceFileFrom(Path...)}</dd>
 *   <dd>{@link #inputSequenceFileFrom(String...)}</dd>
 *   <dd>{@link #inputFileFrom(FileInputFormat,Path...)}</dd>
 *   <dd>{@link #inputFileFrom(FileInputFormat,String...)}</dd>
 *
 *   <dt>Set up the output format</dt>
 *   <dd>{@link #outputTextOn(Path)}</dd>
 *   <dd>{@link #outputTextOn(String)}</dd>
 *   <dd>{@link #outputSequenceFileOn(Path)}</dd>
 *   <dd>{@link #outputSequenceFileOn(String)}</dd>
 *   <dd>{@link #outputFileOn(FileOutputFormat,Path)}</dd>
 *   <dd>{@link #outputFileOn(FileOutputFormat,String)}</dd>
 * </dl>
 *
 * @param <KEYIN>
 *     Input key of the job.
 *
 * @param <VALUEIN>
 *     Input value of the job.
 *
 * @param <KEYMID>
 *     Intermediate key of the job.
 *
 * @param <VALUEMID>
 *     Intermediate value of the job.
 *
 * @param <KEYOUT>
 *     Output key of the job.
 *
 * @param <VALUEOUT>
 *     Output value of the job.
 */
public class JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT > {


    /** Mapper class of the job. */
    private final Class< ? extends Mapper > mapperClass;


    /** Reducer class of the job. */
    private final Class< ? extends Reducer > reducerClass;


    /** Actions that set up a job. */
    private final List< JobConfSetup > setups = new ArrayList< JobConfSetup >();


    /**
     * Constructs a builder with the specified mapper and reducer.
     *
     * <p>This constructor is intended to be accessed from child classes.
     * If you simply want to get a builder instance,
     * use {@linkplain #of(Mapper,Reducer)}.</p>
     *
     * @param mapper
     *     Sample instance of the mapper class of the job.
     *
     * @param reducer
     *     Sample instance of the reducer class of the job.
     *
     * @see #of(Mapper,Reducer)
     */
    protected JobConfBuilder(
            Mapper<KEYIN, VALUEIN, KEYMID, VALUEMID> mapper
            , Reducer<KEYMID, VALUEMID, KEYOUT, VALUEOUT> reducer ) {
        this.mapperClass = mapper.getClass();
        this.reducerClass = reducer.getClass();
    }


    /**
     * Constructs a builder.
     *
     * <p>This constructor is intended to be accessed from child classes.
     * If you simply want to get a builder instance, use {@linkplain #of(Mapper)}.</p>
     *
     * @param mapper
     *     Sample instance of the mapper class of the job.
     */
    protected JobConfBuilder( Mapper<KEYIN, VALUEIN, KEYMID, VALUEMID> mapper ) {
        this.mapperClass = mapper.getClass();
        this.reducerClass = null;
    }


    /**
     * Returns a builder of a job configuration.
     *
     * <h4>Example using generic mapper/reducer classes</h4>
     *
     * <pre>
     * JobConf jobConf = JobConfBuilder.of(
     *         new IdentityMapper&lt; Text , Text &gt;()
     *         , new IdentityReducer&lt; Text , Text &gt;() )
     *     .detectJar()
     *     .outputKeyValue( Text.class , Text.class )
     *     .inputKeyValueTextFrom( "input" )
     *     .outputTextOn( "output" )
     *     .buildJobConf( getConf() );
     * JobClient.runJob( jobConf );
     * </pre>
     *
     * @param mapper
     *     Sample instance of the mapper class of the job.
     *
     * @param reducer
     *     Sample instance of the reducer class of the job.
     *
     * @param <KEYIN>
     *     Input key of the job.
     *
     * @param <VALUEIN>
     *     Input value of the job.
     *
     * @param <KEYMID>
     *     Intermediate key of the job.
     *
     * @param <VALUEMID>
     *     Intermediate value of the job.
     *
     * @param <KEYOUT>
     *     Output key of the job.
     *
     * @param <VALUEOUT>
     *     Output value of the job.
     *
     * @return
     *     Builder of a job configuration.
     */
    public static < KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            of( Mapper<KEYIN, VALUEIN, KEYMID, VALUEMID> mapper
            , Reducer<KEYMID, VALUEMID, KEYOUT, VALUEOUT> reducer ) {
        return new JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            ( mapper , reducer );
    }


    /**
     * Returns a builder of a job configuration with no reducer.
     *
     * <h4>Example using generic a mapper class</h4>
     *
     * <pre>
     * JobConf jobConf = JobConfBuilder.of( new IdentityMapper&lt; Text , Text &gt;() )
     *     .detectJar()
     *     .outputKeyValue( Text.class , Text.class )
     *     .inputKeyValueTextFrom( "input" )
     *     .outputTextOn( "output" )
     *     .buildJobConf( getConf() );
     * JobClient.runJob( jobConf );
     * </pre>
     *
     * @param mapper
     *     Sample instance of the mapper class of the job.
     *
     * @param <KEYIN>
     *     Input key of the job.
     *
     * @param <VALUEIN>
     *     Input value of the job.
     *
     * @param <KEYOUT>
     *     Output key of the job.
     *
     * @param <VALUEOUT>
     *     Output value of the job.
     *
     * @return
     *     Builder of a job configuration.
     */
    public static < KEYIN , VALUEIN , KEYOUT , VALUEOUT >
            JobConfBuilder< KEYIN , VALUEIN , KEYOUT , VALUEOUT , KEYOUT , VALUEOUT >
            of( Mapper<KEYIN, VALUEIN, KEYOUT, VALUEOUT> mapper ) {
        return new JobConfBuilder< KEYIN , VALUEIN , KEYOUT , VALUEOUT , KEYOUT , VALUEOUT >
            ( mapper );
    }


    /**
     * Sets the combiner class of the job.
     *
     * @param combiner
     *     Sample instance of the combiner class of the job.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            combiner( Reducer< KEYMID , VALUEMID , KEYMID , VALUEMID > combiner ) {
        final Class< ? extends Reducer > combinerClass = combiner.getClass();
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setCombinerClass( combinerClass );
            }
        } );
        return this;
    }


    /**
     * Sets the partitioner class of the job.
     *
     * @param partitioner
     *     Sample instance of the partitioner class of the job.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            partitioner( Partitioner< KEYMID , VALUEMID > partitioner ) {
        final Class< ? extends Partitioner > partitionerClass = partitioner.getClass();
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setPartitionerClass( partitionerClass );
            }
        } );
        return this;
    }


    /**
     * Sets the sort comparator class of the job,
     * using {@link JobConf#setOutputKeyComparatorClass}.
     *
     * @param sortComparator
     *     Sample instance of the sort comparator class of the job.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            sortComparator( RawComparator< KEYMID > sortComparator ) {
        final Class< ? extends RawComparator > comparatorClass = sortComparator.getClass();
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setOutputKeyComparatorClass( comparatorClass );
            }
        } );
        return this;
    }


    /**
     * Sets the grouping comparator class of the job,
     * using {@link JobConf#setOutputValueGroupingComparator}.
     *
     * @param groupingComparator
     *     Sample instance of the grouping comparator class of the job.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            groupingComparator( RawComparator< KEYMID > groupingComparator ) {
        final Class< ? extends RawComparator > comparatorClass = groupingComparator.getClass();
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setOutputValueGroupingComparator( comparatorClass );
            }
        } );
        return this;
    }


    /**
     * Sets the intermediate key/value classes of the job.
     *
     * @param mapOutputKeyClass
     *     Intermediate key class of the job.
     *
     * @param mapOutputValueClass
     *     Intermediate value class of the job.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            mapOutputKeyValue( final Class< KEYMID > mapOutputKeyClass
                    , final Class< VALUEMID > mapOutputValueClass ) {
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setMapOutputKeyClass( mapOutputKeyClass );
                jobConf.setMapOutputValueClass( mapOutputValueClass );
            }
        } );
        return this;
    }


    /**
     * Sets the output key/value classes of the job.
     *
     * @param outputKeyClass
     *     Output key class of the job.
     *
     * @param outputValueClass
     *     Output value class of the job.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputKeyValue( final Class< KEYOUT > outputKeyClass
                    , final Class< VALUEOUT > outputValueClass ) {
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setOutputKeyClass( outputKeyClass );
                jobConf.setOutputValueClass( outputValueClass );
            }
        } );
        return this;
    }


    /**
     * Detects and sets the key-value classes.
     *
     * <p>Calling this method is equivalent to calling
     * {@linkplain #detectMapOutputKeyValue}
     * and  {@linkplain #detectOutputKeyValue}.</p>
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            detectKeyValue() {
        detectMapOutputKeyValue();
        detectOutputKeyValue();
        return this;
    }


    /**
     * Detects and sets the output key-value classes.
     *
     * <p>This method detects the types from
     * the type parameters of the output key/value pair of the reducer class.
     * The reducer class must implement {@linkplain Reducer} directly,
     * with raw classes as the type parameters.
     * (It is an implementation limitation. Theoretically, the type inferrence
     * can be much smarter.)</p>
     *
     * <p>If no reducer is specified, or {@linkplain IdentityReducer}
     * is specified, this method detects the types from
     * the type parameters of the output key/value pair of the mapper class.</p>
     *
     * @return
     *     {@code this}.
     *
     * @throws IllegalStateException
     *     When the type inferrence fails.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            detectOutputKeyValue() {
        KeyValue kv = getOutputKeyValue();
        @SuppressWarnings( "unchecked" )
        Class< KEYOUT > key = (Class< KEYOUT >) kv.key;
        @SuppressWarnings( "unchecked" )
        Class< VALUEOUT > value = (Class< VALUEOUT >) kv.value;
        outputKeyValue( key , value );
        return this;
    }


    /**
     * Detects and sets the intermediate key-value classes.
     *
     * <p>This method detects the types from
     * the type parameters of the intermediate key/value pair of the mapper class.
     * The mapper class must implement {@linkplain Mapper} directly,
     * with raw classes as the type parameters.
     * (It is an implementation limitation. Theoretically, the type inferrence
     * can be much smarter.)</p>
     *
     * <p>If {@linkplain IdentityMapper} is specified as the mapper,
     * this method detects the types from
     * the type parameters of the input key/value pair of the reducer class.</p>
     *
     * @return
     *     {@code this}.
     *
     * @throws IllegalStateException
     *     When the type inferrence fails.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            detectMapOutputKeyValue() {
        KeyValue kv = getMapOutputKeyValue();
        @SuppressWarnings( "unchecked" )
        Class< KEYMID > key = (Class< KEYMID >) kv.key;
        @SuppressWarnings( "unchecked" )
        Class< VALUEMID > value = (Class< VALUEMID >) kv.value;
        mapOutputKeyValue( key , value );
        return this;
    }


    /** Gets intermediate key-value classes from the mapper. */
    private KeyValue getMapOutputKeyValue() {
        if ( this.mapperClass.equals( IdentityMapper.class ) ) {
            return getReduceInputKeyValue();
        }
        ParameterizedType mapperInterface
            = getParameterizedInterface( this.mapperClass , Mapper.class );
        Type[] typeArgs = mapperInterface.getActualTypeArguments();
        Type midKey = typeArgs[ 2 ];
        Type midValue = typeArgs[ 3 ];
        Commons.checkClass( midKey , "key" , this.mapperClass );
        Commons.checkClass( midValue , "value" , this.mapperClass );
        return new KeyValue( (Class) midKey , (Class) midValue );
    }


    /** Gets intermediate key-value classes from the reducer. */
    private KeyValue getReduceInputKeyValue() {
        ParameterizedType reducerInterface
            = getParameterizedInterface( this.reducerClass , Reducer.class );
        Type[] typeArgs = reducerInterface.getActualTypeArguments();
        Type midKey = typeArgs[ 0 ];
        Type midValue = typeArgs[ 1 ];
        Commons.checkClass( midKey , "key" , this.reducerClass );
        Commons.checkClass( midValue , "value" , this.reducerClass );
        return new KeyValue( (Class) midKey , (Class) midValue );
    }


    /** Gets output key-value classes. */
    private KeyValue getOutputKeyValue() {
        if ( this.reducerClass == null || this.reducerClass.equals( IdentityReducer.class ) ) {
            return getMapOutputKeyValue();
        }
        ParameterizedType reducerInterface
            = getParameterizedInterface( this.reducerClass , Reducer.class );
        Type[] typeArgs = reducerInterface.getActualTypeArguments();
        Type outKey = typeArgs[ 2 ];
        Type outValue = typeArgs[ 3 ];
        Commons.checkClass( outKey , "key" , this.reducerClass );
        Commons.checkClass( outValue , "value" , this.reducerClass );
        return new KeyValue( (Class) outKey , (Class) outValue );
    }


    /**
     * Returns a parameterized interface of {@code intf} that is implemented by {@code klass}.
     */
    private static ParameterizedType getParameterizedInterface(
            Class < ? > klass , Class< ? > intf ) {
        for ( Type type : klass.getGenericInterfaces() ) {
            if ( Commons.isParameterizedTypeOf( type , intf ) ) {
                return (ParameterizedType) type;
            }
        }
        throw new IllegalStateException(
                "Cannot get generic supertype of " + intf.getName()
                + " from " + klass.getName() );
    }


    /** Pair of key-class and value-class. */
    static final class KeyValue {

        /** Key-class. */
        final Class< ? > key;

        /** Value-class. */
        final Class< ? > value;

        /** Constructs a pair. */
        KeyValue( Class< ? > key , Class< ? > value ) {
            this.key = key;
            this.value = value;
        }
    }


    /**
     * Adds {@code pathName} as a distributed cache file, using
     * {@linkplain DistributedCache#addCacheFile DistributedCache#addCacheFile}.</p>
     *
     * <p>Your mapper and reducer can read the cache file through
     * {@linkplain DistributedCache#getLocalCacheFiles
     * DistributedCache#getLocalCacheFiles}.</p>
     *
     * <h4>Example</h4>
     *
     * <pre>
     * JobConf jobConf = JobConfBuilder.of( new FilterMapper() , new InjectReducer() )
     *     .detectJar()
     *     .detectKeyValue()
     *     .inputKeyValueTextFrom( "input/orders" )
     *     .outputTextOn( "output/statistics" )
     *     .cacheFile( "input/customers.txt" )
     *     .buildJobConf( getConf() );
     * JobClient.runJob( jobConf );
     * </pre>
     *
     * @param pathName
     *     Path of the cache file.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            cacheFile( String pathName ) {
        final URI cacheURI = uri( pathName );
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                DistributedCache.addCacheFile( cacheURI , jobConf );
            }
        } );
        return this;
    }


    /**
     * Adds {@code pathName} as a distributed cache file,
     * and creates a symlink named {@code linkName} in
     * work directories of tasks.
     *
     * <p>Calling this method is equivalent to calling
     * {@code DistributedCache.addCacheFile( new URI( pathName + "#" + linkName , jobConf )}
     * and {@code DistributedCache.createSymlink( jobConf )}.</p>
     *
     * <h4>Example</h4>
     *
     * <pre>
     * JobConf jobConf = JobConfBuilder.of( new FilterMapper() , new InjectReducer() )
     *     .detectJar()
     *     .detectKeyValue()
     *     .inputKeyValueTextFrom( "input/orders" )
     *     .outputTextOn( "output/statistics" )
     *     .cacheFileWithSymlink( "input/customers.txt" , "customers.txt" )
     *     .buildJobConf( getConf() );
     * JobClient.runJob( jobConf );
     *
     * public static class InjectReducer implements Reducer&lt;...&gt; {
     *     private List&lt; String &gt; customers;
     *     &#64;Override public void configure( JobConf jobConf ) {
     *         this.customers = FileUtils.readLines( new File( "customers.txt" ) );
     *     }
     *     &#64;Override public void reduce(...) {
     *         ...
     *     }
     * }
     * </pre>
     *
     * @param pathName
     *     Path of the cache file.
     *
     * @param linkName
     *     Name of the symlink to the cache file.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            cacheFileWithSymlink( String pathName , String linkName ) {
        final URI cacheURI = uri( pathName + "#" + linkName );
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                DistributedCache.addCacheFile( cacheURI , jobConf );
                DistributedCache.createSymlink( jobConf );
            }
        } );
        return this;
    }


    /**
     * Adds {@code pathName} as a distributed cache archive.
     *
     * @param pathName
     *     Path of the cache archive.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            cacheArchive( String pathName ) {
        final URI cacheURI = uri( pathName );
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                DistributedCache.addCacheArchive( cacheURI , jobConf );
            }
        } );
        return this;
    }


    /**
     * Adds {@code pathName} as a distributed cache archive,
     * and creates a symlink named {@code linkName} in
     * work directories of tasktrackers.
     *
     * @param pathName
     *     Path of the cache archive.
     *
     * @param linkName
     *     Name of the symlink to the cache file.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            cacheArchiveWithSymlink( String pathName , String linkName ) {
        final URI cacheURI = uri( pathName + "#" + linkName );
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                DistributedCache.addCacheArchive( cacheURI , jobConf );
                DistributedCache.createSymlink( jobConf );
            }
        } );
        return this;
    }


    /** Returns a uri that represents {@code pathName}. */
    private static URI uri( String pathName ) {
        try {
            return new URI( pathName );
        } catch ( URISyntaxException e ) {
            throw new IllegalArgumentException( "Illegal URI: " + pathName , e );
        }
    }


    /**
     * Uses {@link TextInputFormat} as the input format of the job
     * and sets {@code paths} as the input files/directories.
     *
     * @param paths
     *     Paths from which the job reads texts.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputTextFrom( final Path ... paths ) {
        inputFileFromInternal( TextInputFormat.class , paths );
        return this;
    }


    /**
     * Uses {@link TextInputFormat} as the input format of the job
     * and sets {@code pathNames} as the input files/directories.
     *
     * @param pathNames
     *     Path names from which the job reads texts.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputTextFrom( final String ... pathNames ) {
        inputFileFromInternal( TextInputFormat.class , pathNames );
        return this;
    }


    /**
     * Uses {@link KeyValueTextInputFormat} as the input format of the job
     * and sets {@code paths} as the input files/directories.
     *
     * @param paths
     *     Paths from which the job reads key-value texts.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputKeyValueTextFrom( final Path ... paths ) {
        inputFileFromInternal( KeyValueTextInputFormat.class , paths );
        return this;
    }


    /**
     * Uses {@link KeyValueTextInputFormat} as the input format of the job
     * and sets {@code pathNames} as the input files/directories.
     *
     * @param pathNames
     *     Path names from which the job reads key-value texts.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputKeyValueTextFrom( final String ... pathNames ) {
        inputFileFromInternal( KeyValueTextInputFormat.class , pathNames );
        return this;
    }


    /**
     * Uses {@link SequenceFileInputFormat} as the input format of the job
     * and sets {@code paths} as the input files/directories.
     *
     * @param paths
     *     Paths from which the job reads sequence files.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputSequenceFileFrom( final Path ... paths ) {
        inputFileFromInternal( SequenceFileInputFormat.class , paths );
        return this;
    }


    /**
     * Uses {@link SequenceFileInputFormat} as the input format of the job
     * and sets {@code pathNames} as the input files/directories.
     *
     * @param pathNames
     *     Path names from which the job reads sequence files.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputSequenceFileFrom( final String ... pathNames ) {
        inputFileFromInternal( SequenceFileInputFormat.class , pathNames );
        return this;
    }


    /**
     * Reads input files from {@code paths} using {@code inputFormat}
     * as the input format of the job.
     *
     * @param inputFormat
     *     Class of the input format of the job, which extends {@link FileInputFormat}.
     *
     * @param paths
     *     Paths from which the job reads files.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputFileFrom(
                    FileInputFormat< KEYIN , VALUEIN > inputFormat
                    , Path ... paths ) {
        inputFileFromInternal( inputFormat.getClass() , paths );
        return this;
    }


    /**
     * Reads input files from {@code paths} using {@code inputFormat}
     * as the input format of the job.
     *
     * @param inputFormat
     *     Sample instancef of the input format of the job,
     *     which extends {@link FileInputFormat}.
     *
     * @param pathNames
     *     Paths from which the job reads files.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputFileFrom(
                    FileInputFormat< KEYIN , VALUEIN > inputFormat , String ... pathNames ) {
        inputFileFromInternal( inputFormat.getClass() , pathNames );
        return this;
    }


    /**
     * Reads input files from {@code paths} using {@code inputFormat}
     * as the input format of the job.
     */
    private void inputFileFromInternal(
            final Class< ? extends FileInputFormat > inputFormat
            , final Path ... paths ) {
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                for ( Path path : paths ) {
                    FileInputFormat.addInputPath( jobConf , path );
                }
                jobConf.setInputFormat( inputFormat );
            }
        } );
    }


    /**
     * Reads input files from {@code pathNames} using {@code inputFormat}
     * as the input format of the job.
     */
    private void inputFileFromInternal(
            Class< ? extends FileInputFormat > inputFormat
            , String ... pathNames ) {
        Path[] paths = Commons.pathsFromPathnames( pathNames );
        inputFileFromInternal( inputFormat , paths );
    }


    /**
     * Uses {@link TextOutputFormat} as the output format
     * and sets {@code path} as the output directory.
     *
     * @param path
     *     Path on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputTextOn( final Path path ) {
        outputFileOn( TextOutputFormat.class , path );
        return this;
    }


    /**
     * Uses {@link TextOutputFormat} as the output format
     * and sets {@code pathName} as the output directory.
     *
     * @param pathName
     *     Path name on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputTextOn( String pathName ) {
        return outputTextOn( new Path( pathName ) );
    }


    /**
     * Uses {@link SequenceFileOutputFormat} as the output format
     * and sets {@code path} as the output directory.
     *
     * @param path
     *     Path on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputSequenceFileOn( final Path path ) {
        outputFileOn( SequenceFileOutputFormat.class , path );
        return this;
    }


    /**
     * Uses {@link SequenceFileOutputFormat} as the output format
     * and sets {@code pathName} as the output directory.
     *
     * @param pathName
     *     Path on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputSequenceFileOn( String pathName ) {
        return outputSequenceFileOn( new Path( pathName ) );
    }


    /**
     * Outputs the result to the specified output path
     * using the specified output format.
     *
     * @param outputFormat
     *     Sample instance of the output format of the job,
     *     which extends {@linkplain FileOutputFormat}
     *
     * @param path
     *     Path on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputFileOn( FileOutputFormat< KEYOUT , VALUEOUT > outputFormat , Path path ) {
        outputFileOn( outputFormat.getClass() , path );
        return this;
    }


    /**
     * Outputs the result to the specified output path
     * using the specified output format.
     *
     * @param outputFormat
     *     Sample instance of the output format of the job,
     *     which extends {@linkplain FileOutputFormat}
     *
     * @param pathName
     *     Path on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputFileOn(
                    FileOutputFormat< KEYOUT , VALUEOUT > outputFormat
                    , String pathName ) {
        outputFileOn( outputFormat.getClass() , new Path( pathName ) );
        return this;
    }


    /**
     * Outputs the result to the specified output path
     * using the specified output format.
     *
     * @param outputFormat
     *     Class of the output format of the job,
     *     which extends {@linkplain FileOutputFormat}
     *
     * @param path
     *     Path on which the job outputs the result.
     */
    private void outputFileOn(
                    final Class< ? extends FileOutputFormat > outputFormat
                    , final Path path ) {
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                FileOutputFormat.setOutputPath( jobConf , path );
                jobConf.setOutputFormat( outputFormat );
            }
        } );
    }


    /**
     * Sets the name of the job.
     *
     * @param jobName
     *     Name of the job.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            jobName( final String jobName ) {
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setJobName( jobName );
            }
        } );
        return this;
    }


    /**
     * Sets the number of reduce tasks.
     *
     * @param numReduceTasks
     *     Number of reduce tasks.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            numReduceTasks( final int numReduceTasks ) {
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setNumReduceTasks( numReduceTasks );
            }
        } );
        return this;
    }


    /**
     * Sets the working directory of the job.
     *
     * @param workingDirectory
     *     Path of the working directory.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            workingDirectory( final Path workingDirectory ) {
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setWorkingDirectory( workingDirectory );
            }
        } );
        return this;
    }


    /**
     * Sets the working directory of the job.
     *
     * @param workingDirectoryPathName
     *     Path name of the working directory.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            workingDirectory( String workingDirectoryPathName ) {
        return workingDirectory( new Path( workingDirectoryPathName ) );
    }


    /**
     * Sets the jar file of the job by an example class.
     *
     * @param exampleClass
     *     Class that represents the jar of the job.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            jarByClass( final Class< ? > exampleClass ) {
        addSetup( new JobConfSetup() {
            @Override public void setup( JobConf jobConf ) {
                jobConf.setJarByClass( exampleClass );
            }
        } );
        return this;
    }


    /**
     * Sets the jar file of the job by the class
     * that calls this method.
     *
     * @return
     *     {@code this}.
     */
    public JobConfBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            detectJar() {
        Class< ? > callingClass = Commons.classOnTopOfStack(
                new Throwable().getStackTrace() , getClass() );
        jarByClass( callingClass );
        return this;
    }


    /** Add a set-up action. */
    private void addSetup( JobConfSetup setup ) {
        this.setups.add( setup );
    }


    /**
     * Builds a job configuration.
     *
     * <p>Calling this method is equivalent to create a new {@link JobConf job configuration}
     * with the parent configuration {@code baseConf},
     * and call {@link #setup} with the new job configuration.</p>
     *
     * @param baseConf
     *     Configuration whose settings will be inherited.
     *
     * @return
     *     New job configuration.
     */
    public JobConf buildJobConf( Configuration baseConf ) {
        JobConf jobConf = new JobConf( baseConf );
        setup( jobConf );
        return jobConf;
    }


    /**
     * Sets up the existing job configuration.
     *
     * <p>The following example code is equivalent
     * to <a href="#simplisticexample">the example code on the top of this page</a>.</p>
     *
     * <pre>
     * JobConf jobConf = new JobConf( getConf() );
     * JobConfBuilder.of( new WordCountMapper() , new WordCountReducer() )
     *     .jobName( "word-count" )
     *     .detectJar()
     *     .detectKeyValue()
     *     .combiner( new WordCountReducer() )
     *     .inputTextFrom( "word-count/input" )
     *     .outputSequenceFileOn( "word-count/output" )
     *     .setup( jobConf );
     * JobClient.runJob( jobConf );
     * </pre>
     *
     * @param jobConf
     *     Job configuration to set up.
     */
    public void setup( JobConf jobConf ) {
        jobConf.setMapperClass( this.mapperClass );
        if ( this.reducerClass != null ) {
            jobConf.setReducerClass( this.reducerClass );
        }
        for ( JobConfSetup setup : this.setups ) {
            setup.setup( jobConf );
        }
    }


    /** Set-up action. */
    private interface JobConfSetup {

        /**
         * Does set-up action.
         *
         * @param jobConf
         *     Job configuration to set up.
         */
        void setup( JobConf jobConf );
    }
}

// vim: et sw=4 sts=4
